#include "SDDS.h"
#include "mdb.h"
#include "scan.h"
#if defined(_WIN32)
#  include <io.h>
#  include <fcntl.h>
#endif

#define SET_PIPE 0
#define N_OPTIONS 1

char *option[N_OPTIONS] = {
  "pipe"};

char *USAGE = "sdds2agilentArb [<inputFile>] [<outputFile>] [-pipe[=in][,out]]\n\n\
Converts SDDS to Agilent Arbitrary Waveform files.\n\
Program by Robert Soliday. (" __DATE__ " " __TIME__ ", SVN revision: " SVN_VERSION ")\n";

int main(int argc, char *argv[]) {
  SCANNED_ARG *scanned;
  long iArg;
  unsigned long pipeFlags = 0;
  char *input = NULL, *output = NULL;
  FILE *fd;
  SDDS_DATASET SDDS_dataset;

  long points;
  double *Iwave;
  double *Qwave;
  double maxAmp = 0;
  double minAmp = 0;
  int i;
  double scale;
  short *waveform = NULL;
  char buf;
  char *pChar;

  SDDS_RegisterProgramName(argv[0]);
  argc = scanargs(&scanned, argc, argv);
  if (argc < 2) {
    fprintf(stderr, "%s", USAGE);
    return (1);
  }
  for (iArg = 1; iArg < argc; iArg++) {
    if (scanned[iArg].arg_type == OPTION) {
      switch (match_string(scanned[iArg].list[0], option, N_OPTIONS, 0)) {
      case SET_PIPE:
        if (!processPipeOption(scanned[iArg].list + 1, scanned[iArg].n_items - 1, &pipeFlags)) {
          fprintf(stderr, "invalid -pipe syntax\n");
          return (1);
        }
        break;
      default:
        fprintf(stderr, "invalid option seen\n");
        fprintf(stderr, "%s", USAGE);
        return (1);
      }
    } else {
      if (!input)
        input = scanned[iArg].list[0];
      else if (!output)
        output = scanned[iArg].list[0];
      else {
        fprintf(stderr, "too many filenames\n");
        fprintf(stderr, "%s", USAGE);
        return (1);
      }
    }
  }
  processFilenames("sdds2agilentArb", &input, &output, pipeFlags, 0, NULL);

  if (!SDDS_InitializeInput(&SDDS_dataset, input)) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  }
  if ((SDDS_CheckColumn(&SDDS_dataset, "I", NULL, SDDS_ANY_FLOATING_TYPE, NULL)) != SDDS_CHECK_OKAY) {
    fprintf(stderr, "error: Floating type column named I does not exist\n");
    return (1);
  }
  if ((SDDS_CheckColumn(&SDDS_dataset, "Q", NULL, SDDS_ANY_FLOATING_TYPE, NULL)) != SDDS_CHECK_OKAY) {
    fprintf(stderr, "error: Floating type column named Q does not exist\n");
    return (1);
  }

  if (SDDS_ReadTable(&SDDS_dataset) != 1) {
    fprintf(stderr, "error: No data found in SDDS file\n");
    return (1);
  }
  points = SDDS_RowCount(&SDDS_dataset);

  Iwave = SDDS_GetColumnInDoubles(&SDDS_dataset, "I");
  Qwave = SDDS_GetColumnInDoubles(&SDDS_dataset, "Q");
  if (!SDDS_Terminate(&SDDS_dataset)) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    exit(1);
  }

  maxAmp = Iwave[0];
  minAmp = Iwave[0];
  for (i = 0; i < points; i++) {
    if (maxAmp < Iwave[i])
      maxAmp = Iwave[i];
    else if (minAmp > Iwave[i])
      minAmp = Iwave[i];
    if (maxAmp < Qwave[i])
      maxAmp = Qwave[i];
    else if (minAmp > Qwave[i])
      minAmp = Qwave[i];
  }
  maxAmp = fabs(maxAmp);
  minAmp = fabs(minAmp);
  if (minAmp > maxAmp)
    maxAmp = minAmp;

  scale = 32767 / maxAmp;
  waveform = malloc(sizeof(short) * points * 2);
  for (i = 0; i < points; i++) {
    waveform[2 * i] = (short)floor(Iwave[i] * scale + 0.5);
    waveform[2 * i + 1] = (short)floor(Qwave[i] * scale + 0.5);
  }
  free(Iwave);
  free(Qwave);

  if (!SDDS_IsBigEndianMachine()) {
    pChar = (char *)&waveform[0];
    for (i = 0; i < 2 * points; i++) {
      buf = *pChar;
      *pChar = *(pChar + 1);
      *(pChar + 1) = buf;
      pChar += 2;
    }
  }

  if (!output) {
#if defined(_WIN32)
    if (_setmode(_fileno(stdout), _O_BINARY) == -1) {
      fprintf(stderr, "error: unable to set stdout to binary mode\n");
      exit(1);
    }
#endif
    fd = stdout;
  } else {
    fd = fopen(output, "wb");
  }
  if (fd == NULL) {
    fprintf(stderr, "unable to open output file for writing\n");
    exit(1);
  }
  fwrite((void *)waveform, sizeof(short), points * 2, fd);
  fclose(fd);
  if (waveform)
    free(waveform);
  free_scanargs(&scanned, argc);

  return 0;
}
