/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 * National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
 * This file is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution. 
\*************************************************************************/

#include <math.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_math.h>
#include "SDDS.h"
#include "mdb.h"
#include "scan.h"
#include "SDDSutils.h"

double bandwidth(double data[], int64_t M);
double gaussiankernelfunction(double sample);
double kerneldensityestimate(double *trainingdata_x, double *trainingdata_y,
                             double sample_x, double sample_y, int64_t n);
double *gridX(double start, double end, int N);
double *gridY(double start, double end, int N);

#define SET_COLUMN 0
#define SET_PIPE 1
#define SET_MARGIN 2
#define SET_SAME_SCALES 3
#define N_OPTIONS 4

char *option[N_OPTIONS] = {
  "column",
  "pipe",
  "margin",
  "samescales"};

static char *USAGE = "sddskde2d [<inputfile>] [<outputfile>] [-column=<2 columns>] [-samescales] [-pipe] [-margin=<value>]\n\
-column         Provide 2 column names separate by comma, wild card accepted.\n\
-margin         Provide the ratio to extend the original data, default 0.05. \n\
-sameScales     Use the same x and y ranges for all pages of output.\n\
-pipe           The standard SDDS Toolkit pipe option.\n\n\
sddskde2d does kernel density estimation for two dimensional data.\n\
Program by Yipeng Sun (" __DATE__ " " __TIME__ ", SVN revision: " SVN_VERSION ")\n";

int main(int argc, char **argv) {
  int n_test, n_total, sameScales;
  double lowerx, upperx;
  double lowery, uppery;
  double margin = 0.05;
  SDDS_DATASET SDDSin, SDDSout;
  char *inputFile, *outputFile, **column;
  long tmpfile_used = 0, columns, noWarnings = 1;
  unsigned long pipeFlags;
  long i, i_arg, j;
  SCANNED_ARG *s_arg;
  double **columnDataX, **columnDataY, *pdf;
  int64_t *rows, rows0;
  int32_t nPages, iPage;

  inputFile = outputFile = NULL;
  pdf = NULL;
  columns = 0;
  column = NULL;
  sameScales = 0;

  n_test = 50;
  n_total = n_test * n_test;

  SDDS_RegisterProgramName(argv[0]);
  argc = scanargs(&s_arg, argc, argv);
  pipeFlags = 0;
  tmpfile_used = 0;
  if (argc < 2) {
    fprintf(stderr, "%s", USAGE);
    exit(1);
  }
  for (i_arg = 1; i_arg < argc; i_arg++) {
    if (s_arg[i_arg].arg_type == OPTION) {
      delete_chars(s_arg[i_arg].list[0], "_");
      switch (match_string(s_arg[i_arg].list[0], option, N_OPTIONS, 0)) {
      case SET_COLUMN:
        columns = s_arg[i_arg].n_items - 1;
        column = trealloc(column, sizeof(*column) * columns);
        for (i = 1; i < s_arg[i_arg].n_items; i++)
          column[i - 1] = s_arg[i_arg].list[i];
        break;
      case SET_MARGIN:
        if (s_arg[i_arg].n_items != 2)
          SDDS_Bomb("Invalid -margin option. Too many qualifiers.");
        if (!get_double(&margin, s_arg[i_arg].list[1]))
          SDDS_Bomb("Invalid -margin value provided.");
        break;
      case SET_SAME_SCALES:
        if (s_arg[i_arg].n_items != 1)
          SDDS_Bomb("Invalid -sameScales option. No qualifiers are accepted.");
        sameScales = 1;
        break;
      case SET_PIPE:
        if (!processPipeOption(s_arg[i_arg].list + 1, s_arg[i_arg].n_items - 1, &pipeFlags)) {
          fprintf(stderr, "Error (%s): invalid -pipe syntax\n", argv[0]);
          return (1);
        }
        break;
      }
    } else {
      if (!inputFile)
        inputFile = s_arg[i_arg].list[0];
      else if (outputFile == NULL)
        outputFile = s_arg[i_arg].list[0];
      else {
        fprintf(stderr, "Error (%s): too many filenames\n", argv[0]);
        return (1);
      }
    }
  }
  processFilenames("sddskde2d", &inputFile, &outputFile, pipeFlags, noWarnings, &tmpfile_used);
  if (!columns) {
    fprintf(stderr, "%s", USAGE);
    SDDS_Bomb("No column provided!");
  }
  if (!SDDS_InitializeInput(&SDDSin, inputFile)) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return (1);
  }
  if ((columns = expandColumnPairNames(&SDDSin, &column, NULL, columns, NULL, 0, FIND_NUMERIC_TYPE, 0)) <= 0) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    SDDS_Bomb("no columns selected.");
  }
  if (columns > 2) {
    fprintf(stderr, "%s", USAGE);
    SDDS_Bomb("Only 2 columns may be accepted.");
  }
  if (!SDDS_InitializeOutput(&SDDSout, SDDS_BINARY, 1, NULL, NULL, outputFile))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (!SDDS_TransferColumnDefinition(&SDDSout, &SDDSin, column[0], NULL) ||
      !SDDS_TransferColumnDefinition(&SDDSout, &SDDSin, column[1], NULL) ||
      SDDS_DefineColumn(&SDDSout, "PDF", NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0) < 0)
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (!SDDS_WriteLayout(&SDDSout))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

  /* Read all the data in case we need to use fixed scales */
  nPages = 0;
  columnDataX = NULL;
  columnDataY = NULL;
  rows = NULL;
  while (SDDS_ReadPage(&SDDSin) >= 0) {
    if ((rows0 = SDDS_CountRowsOfInterest(&SDDSin))) {
      columnDataX = (double **)SDDS_Realloc(columnDataX, sizeof(*columnDataX) * (nPages + 1));
      columnDataY = (double **)SDDS_Realloc(columnDataY, sizeof(*columnDataY) * (nPages + 1));
      rows = (int64_t *)SDDS_Realloc(rows, sizeof(*rows) * (nPages + 1));
      rows[nPages] = rows0;
      if (!(columnDataX[nPages] = SDDS_GetColumnInDoubles(&SDDSin, column[0])))
        SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
      if (!(columnDataY[nPages] = SDDS_GetColumnInDoubles(&SDDSin, column[1])))
        SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
      nPages++;
    }
  }
  if (nPages == 0)
    SDDS_Bomb("No data in file");

  if (sameScales) {
    double llx, lly, uux, uuy;
    lowerx = lowery = DBL_MAX;
    upperx = uppery = -DBL_MAX;
    for (iPage = 0; iPage < nPages; iPage++) {
      find_min_max(&llx, &uux, columnDataX[iPage], rows[iPage]);
      lowerx = MIN(llx, lowerx);
      upperx = MAX(uux, upperx);
      find_min_max(&lly, &uuy, columnDataY[iPage], rows[iPage]);
      lowery = MIN(lly, lowery);
      uppery = MAX(uuy, uppery);
    }
  }

  pdf = malloc(sizeof(*pdf) * n_total);
  for (iPage = 0; iPage < nPages; iPage++) {
    if (!SDDS_StartPage(&SDDSout, n_total))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    if (!sameScales) {
      find_min_max(&lowerx, &upperx, columnDataX[iPage], rows[iPage]);
      find_min_max(&lowery, &uppery, columnDataY[iPage], rows[iPage]);
    }
    double *x_array = gridX(lowerx, upperx, n_test);
    double *y_array = gridY(lowery, uppery, n_test);
    for (j = 0; j < n_total; j++) {
      pdf[j] = kerneldensityestimate(columnDataX[iPage], columnDataY[iPage], x_array[j], y_array[j], rows[iPage]);
    }
    if (!SDDS_SetColumnFromDoubles(&SDDSout, SDDS_SET_BY_NAME, x_array, n_total, column[0]))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    if (!SDDS_SetColumnFromDoubles(&SDDSout, SDDS_SET_BY_NAME, y_array, n_total, column[1]))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    if (!SDDS_SetColumnFromDoubles(&SDDSout, SDDS_SET_BY_NAME, pdf, n_total, "PDF"))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    free(columnDataX[iPage]);
    free(columnDataY[iPage]);
    free(x_array);
    free(y_array);
    if (!SDDS_WritePage(&SDDSout))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  }
  free(pdf);
  free(columnDataX);
  free(columnDataY);
  free(rows);
  if (!SDDS_Terminate(&SDDSin) || !SDDS_Terminate(&SDDSout))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (tmpfile_used && !replaceFileAndBackUp(inputFile, outputFile)) {
    return (1);
  }
  free(column);
  /* free_scanargs(&s_arg,argc); */
  return 0;
}

double bandwidth(double data[], int64_t M) {
  double sigma, bwidth, silver_factor;
  sigma = gsl_stats_sd(data, 1, M);
  silver_factor = pow(M, -0.16666666);
  bwidth = pow(silver_factor, 2.0) * pow(sigma, 2.0);
  return bwidth;
}

double gaussiankernelfunction(double sample) {
  double k;
  k = exp(-sample / 2.0);
  k = k / (2.0 * M_PI);
  return k;
}

double kerneldensityestimate(double *trainingdata_x, double *trainingdata_y,
                             double sample_x, double sample_y, int64_t n) {
  int64_t i;
  double pdf, hx, hy, z;

  hx = bandwidth(trainingdata_x, n);
  hy = bandwidth(trainingdata_y, n);
  pdf = 0.0;
  for (i = 0; i < n; i++) {
    z = (trainingdata_x[i] - sample_x) * (trainingdata_x[i] - sample_x) / hx;
    z += (trainingdata_y[i] - sample_y) * (trainingdata_y[i] - sample_y) / hy;
    pdf += gaussiankernelfunction(z);
  }
  pdf = pdf / (n * sqrt(hx) * sqrt(hy));
  return pdf;
}

double *gridX(double start, double end, int N) {
  double *x;
  int i, j, nGrid;
  double step;

  nGrid = N * N;
  x = (double *)calloc(nGrid, sizeof(double));
  step = (end - start) / (double)(N - 1);

  i = 0;
  for (j = 0; j < N; j++) {
    for (i = 0; i < N; i++) {
      x[i + j * N] = start + i * step;
    }
  }
  return x;
}

double *gridY(double start, double end, int N) {
  double *x;
  int i, j, nGrid;
  double step;

  nGrid = N * N;
  x = (double *)calloc(nGrid, sizeof(double));
  step = (end - start) / (double)(N - 1);

  i = 0;
  for (j = 0; j < N; j++) {
    for (i = 0; i < N; i++) {
      x[i + j * N] = start + j * step;
    }
  }
  return x;
}
