/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 * National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
 * This file is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution. 
\*************************************************************************/

#include "mdb.h"
#include "scan.h"
#include "match_string.h"
#include "SDDS.h"

#define CLO_COLUMNS 0
#define CLO_VERBOSE 1
#define CLO_ASCII 2
#define CLO_PIPE 3
#define CLO_MAJOR_ORDER 4
#define CLO_COMMUTATION_MODE 5
#define CLO_FHEAD 6
#define CLO_REMOVE_COMMUTATION_OFFSET_ONLY 7
#define COMMANDLINE_OPTIONS 8
char *commandline_option[COMMANDLINE_OPTIONS] = {
  "columns",
  "verbose",
  "ascii",
  "pipe",
  "majorOrder",
  "commutationMode",
  "fhead",
  "removeCommutationOffsetOnly",
};

static char *USAGE = "sddsremoveoffsets <input-file> <output-file> [-pipe=[input],[output]] \n\
 -columns=<name> \n\
 [-commutationMode=<string>] \n\
 [-removeCommutationOffsetOnly] \n\
 [-fhead=<value>] \n\
 [-majorOrder=row|column] \n\
 -verbose \n\n\
Removes offset from bpm waveform data with \"commutation\" on. \n\
The resulting data of the resulting file will have zero average. \n\
columns             specifies the data in the input file to be adjusted.\n\
pipe                writes the output to a pipe.\n\
commutationMode     mode to control the removal of offsets using commutation\n\
                    mode codes a, b, ab1, and ab2. \n\
                    a and b means no switching of offsets (use only one offset value)\n\
                    ab1 uses two offsets, one for each alternating rows, \n\
                    ab2 uses two offsets, one for each alternating pairs of rows. \n\
fhead               fraction of rows at the head which will be used to determine the offsets. \n\
removeCommutationOffsetOnly removes only the commutation offset, leaving the average offset.\n\
majorOrder          specifies the output file in row or column major order.\n\
Program by Louis Emery and Jialun Luo, ANL (" __DATE__ " " __TIME__ ", SVN revision: " SVN_VERSION ")\n";

static long commutationModes = 4;
static char *commutationModeList[4] = {"a", "b", "ab1", "ab2"};

static long patternLength = 4;

long resolveColumnNames(SDDS_DATASET *SDDSin, char ***column, int32_t *columns);

int main(int argc, char **argv) {
  SCANNED_ARG *s_arg;
  SDDS_DATASET SDDSin, SDDSout;
  unsigned long majorOrderFlag;
  char *input, *output;
  int64_t j, rows;
  int32_t columns = 0;
  char **inputColumn, **outputColumn;

  long i, i_arg;
  long verbose;
  long commutationMode;
  long removeCommutationOffsetOnly = 0;
  unsigned long pipeFlags;
  long tmpfile_used, noWarnings;
  short columnMajorOrder = -1;
  double *data, offset1, offset2, aveOffset;
  double fhead = 1;

  input = output = NULL;
  inputColumn = outputColumn = NULL;

  SDDS_RegisterProgramName(argv[0]);
  argc = scanargs(&s_arg, argc, argv);
  if (argc == 1)
    bomb(NULL, USAGE);

  verbose = 0;
  input = output = NULL;
  noWarnings = 0;
  pipeFlags = 0;
  tmpfile_used = 0;
  aveOffset = offset1 = offset2 = 0;

  for (i_arg = 1; i_arg < argc; i_arg++) {
    if (s_arg[i_arg].arg_type == OPTION) {
      switch (match_string(s_arg[i_arg].list[0], commandline_option, COMMANDLINE_OPTIONS, UNIQUE_MATCH)) {
      case CLO_MAJOR_ORDER:
        majorOrderFlag = 0;
        s_arg[i_arg].n_items--;
        if (s_arg[i_arg].n_items > 0 && (!scanItemList(&majorOrderFlag, s_arg[i_arg].list + 1, &s_arg[i_arg].n_items, 0, "row", -1, NULL, 0, SDDS_ROW_MAJOR_ORDER, "column", -1, NULL, 0, SDDS_COLUMN_MAJOR_ORDER, NULL)))
          SDDS_Bomb("invalid -majorOrder syntax/values");
        if (majorOrderFlag & SDDS_COLUMN_MAJOR_ORDER)
          columnMajorOrder = 1;
        else if (majorOrderFlag & SDDS_ROW_MAJOR_ORDER)
          columnMajorOrder = 0;
        break;
      case CLO_COMMUTATION_MODE:
        if (s_arg[i_arg].n_items != 2)
          bomb("invalid -commutationMode syntax", NULL);
        if ((commutationMode = match_string(str_tolower(s_arg[i_arg].list[1]), commutationModeList, commutationModes, EXACT_MATCH)) < 0)
          SDDS_Bomb("invalid commutationMode given!");
        break;
      case CLO_VERBOSE:
        verbose = 1;
        break;
      case CLO_REMOVE_COMMUTATION_OFFSET_ONLY:
        removeCommutationOffsetOnly = 1;
        break;
      case CLO_FHEAD:
        if (s_arg[i_arg].n_items != 2)
          bomb("invalid -fhead syntax", NULL);
        if (sscanf(s_arg[i_arg].list[1], "%lf", &fhead) != 1 || fhead <= 0 || fhead > 1)
          fprintf(stderr, "Error: invalid -fhead syntax\n");
        break;
      case CLO_PIPE:
        if (!processPipeOption(s_arg[i_arg].list + 1, s_arg[i_arg].n_items - 1, &pipeFlags))
          SDDS_Bomb("invalid -pipe syntax");
        break;
      case CLO_COLUMNS:
        if (columns)
          SDDS_Bomb("only one -columns option may be given");
        if (s_arg[i_arg].n_items < 2)
          SDDS_Bomb("invalid -columns syntax");
        inputColumn = tmalloc(sizeof(*inputColumn) * (columns = s_arg[i_arg].n_items - 1));
        for (i = 0; i < columns; i++)
          inputColumn[i] = s_arg[i_arg].list[i + 1];
        break;
      default:
        SDDS_Bomb("unrecognized option given");
      }
    } else {
      if (!input)
        input = s_arg[i_arg].list[0];
      else if (!output)
        output = s_arg[i_arg].list[0];
      else
        SDDS_Bomb("too many filenames seen");
    }
  }

  processFilenames("sddsremoveoffsets", &input, &output, pipeFlags, noWarnings, &tmpfile_used);

  if (!columns)
    SDDS_Bomb("supply the names of columns for offset removal with the -columns option");

  if (!SDDS_InitializeInput(&SDDSin, input))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

  /* expand the list of columns from wildcards */
  if (!resolveColumnNames(&SDDSin, &inputColumn, &columns))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (!columns)
    SDDS_Bomb("no columns selected for offset removal");

  if (!SDDS_InitializeCopy(&SDDSout, &SDDSin, output, "w"))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

  outputColumn = tmalloc(sizeof(*outputColumn) * columns);
  for (i = 0; i < columns; i++)
    outputColumn[i] = inputColumn[i];

  if (!SDDS_InitializeOutput(&SDDSout, SDDS_BINARY, 1, "Remove offset from BPM wabeforms", "Modified data", output) ||
      !SDDS_InitializeCopy(&SDDSout, &SDDSin, output, "w"))
    SDDS_PrintErrors(stdout, SDDS_EXIT_PrintErrors | SDDS_VERBOSE_PrintErrors);
  if (columnMajorOrder != -1)
    SDDSout.layout.data_mode.column_major = columnMajorOrder;
  else
    SDDSout.layout.data_mode.column_major = SDDSin.layout.data_mode.column_major;
  if (!SDDS_WriteLayout(&SDDSout))
    SDDS_PrintErrors(stdout, SDDS_EXIT_PrintErrors | SDDS_VERBOSE_PrintErrors);

  while (0 < SDDS_ReadPage(&SDDSin)) {
    if (!SDDS_CopyPage(&SDDSout, &SDDSin))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    if ((rows = SDDS_CountRowsOfInterest(&SDDSin))) {
      for (i = 0; i < columns; i++) {
        if (!(data = SDDS_GetColumnInDoubles(&SDDSin, inputColumn[i])))
          SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
        /* do something with array data */
        long repeatOffset;
        if (data[0] == data[1])
          repeatOffset = 0;
        else
          repeatOffset = 1;

        double sum1;
        double sum2;
        unsigned long count1;
        unsigned long count2;

        sum1 = sum2 = 0;
        count1 = count2 = 0;
        unsigned long halfPatternLength = patternLength / 2;
        for (j = 0; j < fhead * rows; j++) {
          /* printf("%ld\n", j % patternLength); */
          if ((j - repeatOffset) % patternLength < halfPatternLength) {
            sum1 += data[j];
            count1++;
            /* printf("%ld\n", count1); */
          } else {
            sum2 += data[j];
            count2++;
          }
        }
        offset1 = sum1 / count1;
        offset2 = sum2 / count2;
        if (!removeCommutationOffsetOnly) {
          for (j = 0; j < rows; j++) {
            if ((j - repeatOffset) % patternLength < halfPatternLength)
              data[j] = data[j] - offset1;
            else
              data[j] = data[j] - offset2;
          }
        } else {
          aveOffset = (offset1 + offset2) / 2.0;
          for (j = 0; j < rows; j++) {
            if ((j - repeatOffset) % patternLength < halfPatternLength)
              data[j] = data[j] - offset1 + aveOffset;
            else
              data[j] = data[j] - offset2 + aveOffset;
          }
        }

        if (verbose) {
          printf("offset1 = %f \t offset2 = %f\n", offset1, offset2);
          double newMean = 0;
          double newSum = 0;
          for (j = 0; j < rows; j++) {
            newSum += data[j];
          }
          newMean = newSum / (rows + 1);
          printf("New average: %f\n", newMean);
        }

        if (!SDDS_SetColumnFromDoubles(&SDDSout, SDDS_SET_BY_NAME, data, rows, outputColumn[i]))
          SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
        free(data);
      }
    }
    if (!SDDS_WritePage(&SDDSout))
      SDDS_PrintErrors(stdout, SDDS_EXIT_PrintErrors | SDDS_VERBOSE_PrintErrors);
  }

  if (!SDDS_Terminate(&SDDSin) || !SDDS_Terminate(&SDDSout))
    SDDS_PrintErrors(stdout, SDDS_EXIT_PrintErrors | SDDS_VERBOSE_PrintErrors);

  if (tmpfile_used && !replaceFileAndBackUp(input, output))
    exit(1);

  return (0);
}

long resolveColumnNames(SDDS_DATASET *SDDSin, char ***column, int32_t *columns) {
  long i;

  SDDS_SetColumnFlags(SDDSin, 0);
  for (i = 0; i < *columns; i++) {
    if (!SDDS_SetColumnsOfInterest(SDDSin, SDDS_MATCH_STRING, (*column)[i], SDDS_OR))
      return 0;
  }
  free(*column);
  if (!(*column = SDDS_GetColumnNames(SDDSin, columns)) || *columns == 0) {
    SDDS_SetError("no columns found");
    return 0;
  }
  return 1;
}
