/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 * National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
 * This file is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution. 
\*************************************************************************/

/* file: SDDS_info.c
 * contents: user's routines for getting information about definitions.
 * 
 * Michael Borland, November 1993
 $Log: not supported by cvs2svn $
 Revision 1.5  2002/08/14 16:34:48  soliday
 Added Open License

 Revision 1.4  1996/03/21 16:38:35  borland
 Major revision.
 Added buffered binary reading (SDDS_binary.c).
 Added sparsed table reading (SDDS_binary.c, SDDS_ascii.c).
 Added sorted column, parameter, and array name lists (SDDS_info.c,
 SDDS_output.c, SDDS_utils.c).

 * Revision 1.3  1996/03/12  04:00:45  borland
 * Reformatted code to 2-space indentation in K&R style.
 *
 * Revision 1.2  1995/09/05  21:10:57  saunders
 * First test release of the SDDS1.5 package.
 *
 */
/** @file */

/** @mainpage SDDSlib Documentation
 *
 * These library functions are used to read and write <a href="https://ops.aps.anl.gov/SDDSInfo.shtml">SDDS</a> files in C/C++.\n
 * Find all the documented functions <a href="https://ops.aps.anl.gov/manuals/SDDSlib/html/globals.html">here</a>\n
 * Find all the documented and undocumented functions, as well as data structures at: @link SDDS.h @endlink
 *
 * @section examples Code Examples
 * @subsection reading_sdds Accessing Data Stored in an SDDS File
 *
\code{.c}
SDDS_DATASET SDDS_dataset;  

// open the file and read the SDDS header
SDDS_InitializeInput(&SDDS_dataset, filename); 

// see if parameters that are needed are present
if ((SDDS_CheckParameter(&SDDS_dataset, parameter_I_need, NULL, SDDS_ANY_NUMERIC_TYPE, stderr)) != SDDS_CHECK_OKAY) {  
    fprintf(stderr, "parameter %s is not in the data file", parameter_I_need);  
    exit(1);  
}  
 
// see if columns that are needed are present
if ((SDDS_CheckColumn(&SDDS_dataset, column_I_need, NULL, SDDS_ANY_NUMERIC_TYPE, stderr)) != SDDS_CHECK_OKAY) {  
    fprintf(stderr, "column %s is not in the data file", column_I_need);  
    exit(1);  
}  

// read and process each data table in the data set
while (SDDS_ReadPage(&SDDS_dataset)>0) {  
    // access parameter data
    SDDS_GetParameter(...) 
 
    // access column data
    SDDS_GetColumn(...)  
}

// close the file
SDDS_Terminate(&SDDS_dataset); 
\endcode
 *
 * @subsection writing_sdds SDDS Output of Internally Generated Data
 *
\code{.c}
SDDS_DATASET SDDS_dataset;  
 
// open the file and set a few data set properites
SDDS_InitializeOutput(&SDDS_dataset, ...)  
 
// define columns 
SDDS_DefineColumn(&SDDS_dataset, column_name, ...);  
 
// define arrays 
SDDS_DefineArray(&SDDS_dataset, array_name, ...);  
 
// define parameters 
SDDS_DefineParameter(&SDDS_dataset, parameter_name, ...);  
 
// save the header 
SDDS_SaveLayout(&SDDS_dataset);  
 
// write the header 
SDDS_WriteLayout(&SDDS_dataset);  
 
// generate and output the data 
while (DataGenerated(my_data_structure)) {  
    // start a new SDDS data table 
    SDDS_StartPage(&SDDS_dataset, rows);  
 
    // put data into SDDS parameters and columns 
    SDDS_SetParameters(&SDDS_dataset, ...);  

    SDDS_SetArray(&SDDS_dataset, ...);  

    SDDS_SetColumn(&SDDS_dataset, ...);  
 
    // write the new data table 
    SDDS_WritePage(&SDDS_dataset);  
}  
 
SDDS_Terminate(&SDDS_dataset); 
\endcode
 *
 */

#include "mdb.h"
#include "match_string.h"
#include "SDDS.h"
#include "SDDS_internal.h"

/**
 * Returns information about a specified column. This routine is the preferred alternative to SDDS_GetColumnDefinition.
 *
 * @param SDDS_dataset: Address of the SDDS_DATASET structure for the data set.
 * @param field_name: A NULL terminated string giving the name of the field that information is requested for.
 * @param memory: The address in which to place the information. Should have type data-type* in the user's program, where data-type is the data type of the information. Note that for STRING information, data-type is char *. If memory is NULL, the existence and type of the information will be verified; the type will be returned as usual.
 * @param mode: Valid values are SDDS_GET_BY_NAME and SDDS_GET_BY_INDEX\n
                SDDS_GET_BY_NAME - SDDS_GetColumnInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, char *column_name)\n
                SDDS_GET_BY_INDEX - SDDS_GetColumnInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, int32_t column_index)
 *
 * @return On success, returns the SDDS data type of the information. On failure, returns zero and records an error message.
 *
 */
int32_t SDDS_GetColumnInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, ...) {
  int32_t field_index, type;
  int32_t column_index;
  COLUMN_DEFINITION *columndef;
  char *column_name;
  va_list argptr;
  int32_t retval;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_GetColumnInformation"))
    return (0);

  if (!field_name) {
    SDDS_SetError("NULL field name passed. (SDDS_GetColumnInformation)");
    return (0);
  }

  va_start(argptr, mode);
  retval = 1;
  if (mode & SDDS_GET_BY_INDEX) {
    if ((column_index = va_arg(argptr, int32_t)) < 0 || column_index >= SDDS_dataset->layout.n_columns) {
      SDDS_SetError("Invalid column index passed. (SDDS_GetColumnInformation)");
      retval = 0;
    }
  } else {
    if (!(column_name = va_arg(argptr, char *))) {
      SDDS_SetError("NULL column name passed. (SDDS_GetColumnInformation)");
      retval = 0;
    }
    if ((column_index = SDDS_GetColumnIndex(SDDS_dataset, column_name)) < 0) {
      SDDS_SetError("Unknown column name given (SDDS_GetColumnInformation)");
      retval = 0;
    }
  }
  columndef = SDDS_dataset->layout.column_definition + column_index;
  va_end(argptr);
  if (!retval)
    return (0);

  for (field_index = 0; field_index < SDDS_COLUMN_FIELDS; field_index++)
    if (strcmp(field_name, SDDS_ColumnFieldInformation[field_index].name) == 0)
      break;
  if (field_index == SDDS_COLUMN_FIELDS) {
    SDDS_SetError("Unknown field name given (SDDS_GetColumnInformation)");
    return (0);
  }
  type = SDDS_ColumnFieldInformation[field_index].type;
  if (!memory)
    return (type);
  if (type == SDDS_STRING) {
    if (!SDDS_CopyString((char **)memory, *((char **)((char *)columndef + SDDS_ColumnFieldInformation[field_index].offset)))) {
      SDDS_SetError("Unable to copy field data (SDDS_GetColumnInformation)");
      return (0);
    }
  } else
    memcpy(memory, (char *)columndef + SDDS_ColumnFieldInformation[field_index].offset, SDDS_type_size[type - 1]);
  return (type);
}

/**
 * Returns information about a specified parameter. This routine is the preferred alternative to SDDS_GetParameterDefinition.
 *
 * @param SDDS_dataset: Address of the SDDS_DATASET structure for the data set.
 * @param field_name: A NULL terminated string giving the name of the field that information is requested for.
 * @param memory: The address in which to place the information. Should have type data-type* in the user's program, where data-type is the data type of the information. Note that for STRING information, data-type is char *. If memory is NULL, the existence and type of the information will be verified; the type will be returned as usual.
 * @param mode: Valid values are SDDS_GET_BY_NAME and SDDS_GET_BY_INDEX\n
                SDDS_GET_BY_NAME - SDDS_GetParameterInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, char *parameter_name)\n
                SDDS_GET_BY_INDEX - SDDS_GetParameterInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, int32_t parameter_index)
 *
 * @return On success, returns the SDDS data type of the information. On failure, returns zero and records an error message.
 *
 */
int32_t SDDS_GetParameterInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, ...) {
  int32_t field_index, type, parameter_index;
  PARAMETER_DEFINITION *parameterdef;
  char *parameter_name;
  va_list argptr;
  int32_t retval;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_GetParameterInformation"))
    return (0);

  if (!field_name) {
    SDDS_SetError("NULL field name passed. (SDDS_GetParameterInformation)");
    return (0);
  }

  va_start(argptr, mode);
  retval = 1;
  if (mode & SDDS_GET_BY_INDEX) {
    if ((parameter_index = va_arg(argptr, int32_t)) < 0 || parameter_index >= SDDS_dataset->layout.n_parameters) {
      SDDS_SetError("Invalid parameter index passed. (SDDS_GetParameterInformation)");
      retval = 0;
    }
  } else {
    if (!(parameter_name = va_arg(argptr, char *))) {
      SDDS_SetError("NULL parameter name passed. (SDDS_GetParameterInformation)");
      retval = 0;
    }
    if ((parameter_index = SDDS_GetParameterIndex(SDDS_dataset, parameter_name)) < 0) {
      SDDS_SetError("Unknown parameter name given (SDDS_GetParameterInformation)");
      retval = 0;
    }
  }
  parameterdef = SDDS_dataset->layout.parameter_definition + parameter_index;
  va_end(argptr);
  if (!retval)
    return (0);

  for (field_index = 0; field_index < SDDS_PARAMETER_FIELDS; field_index++)
    if (strcmp(field_name, SDDS_ParameterFieldInformation[field_index].name) == 0)
      break;
  if (field_index == SDDS_PARAMETER_FIELDS) {
    SDDS_SetError("Unknown field name given (SDDS_GetParameterInformation)");
    return (0);
  }
  type = SDDS_ParameterFieldInformation[field_index].type;
  if (!memory)
    return (type);
  if (type == SDDS_STRING) {
    if (!SDDS_CopyString((char **)memory, *((char **)((char *)parameterdef + SDDS_ParameterFieldInformation[field_index].offset)))) {
      SDDS_SetError("Unable to copy field data (SDDS_GetParameterInformation)");
      return (0);
    }
  } else
    memcpy(memory, (char *)parameterdef + SDDS_ParameterFieldInformation[field_index].offset, SDDS_type_size[type - 1]);
  return (type);
}

/**
 * Returns information about a specified array. This routine is the preferred alternative to SDDS_GetArrayDefinition.
 *
 * @param SDDS_dataset: Address of the SDDS_DATASET structure for the data set.
 * @param field_name: A NULL terminated string giving the name of the field that information is requested for.
 * @param memory: The address in which to place the information. Should have type data-type* in the user's program, where data-type is the data type of the information. Note that for STRING information, data-type is char *. If memory is NULL, the existence and type of the information will be verified; the type will be returned as usual.
 * @param mode: Valid values are SDDS_GET_BY_NAME and SDDS_GET_BY_INDEX\n
                SDDS_GET_BY_NAME - SDDS_GetArrayInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, char *array_name)\n
                SDDS_GET_BY_INDEX - SDDS_GetArrayInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, int32_t array_index)
 *
 * @return On success, returns the SDDS data type of the information. On failure, returns zero and records an error message.
 *
 */
int32_t SDDS_GetArrayInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, ...) {
  int32_t field_index, type, array_index;
  ARRAY_DEFINITION *arraydef;
  char *array_name;
  va_list argptr;
  int32_t retval;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_GetArrayInformation"))
    return (0);

  if (!field_name) {
    SDDS_SetError("NULL field name passed. (SDDS_GetArrayInformation)");
    return (0);
  }

  va_start(argptr, mode);
  retval = 1;
  if (mode & SDDS_GET_BY_INDEX) {
    if ((array_index = va_arg(argptr, int32_t)) < 0 || array_index >= SDDS_dataset->layout.n_arrays) {
      SDDS_SetError("Invalid array index passed. (SDDS_GetArrayInformation)");
      retval = 0;
    }
  } else {
    if (!(array_name = va_arg(argptr, char *))) {
      SDDS_SetError("NULL array name passed. (SDDS_GetArrayInformation)");
      retval = 0;
    }
    if ((array_index = SDDS_GetArrayIndex(SDDS_dataset, array_name)) < 0) {
      SDDS_SetError("Unknown array name given (SDDS_GetArrayInformation)");
      retval = 0;
    }
  }
  arraydef = SDDS_dataset->layout.array_definition + array_index;
  va_end(argptr);
  if (!retval)
    return (0);

  for (field_index = 0; field_index < SDDS_ARRAY_FIELDS; field_index++)
    if (strcmp(field_name, SDDS_ArrayFieldInformation[field_index].name) == 0)
      break;
  if (field_index == SDDS_ARRAY_FIELDS) {
    SDDS_SetError("Unknown field name given (SDDS_GetArrayInformation)");
    return (0);
  }
  type = SDDS_ArrayFieldInformation[field_index].type;
  if (!memory)
    return (type);
  if (type == SDDS_STRING) {
    if (!SDDS_CopyString((char **)memory, *((char **)((char *)arraydef + SDDS_ArrayFieldInformation[field_index].offset)))) {
      SDDS_SetError("Unable to copy field data (SDDS_GetArrayInformation)");
      return (0);
    }
  } else
    memcpy(memory, (char *)arraydef + SDDS_ArrayFieldInformation[field_index].offset, SDDS_type_size[type - 1]);
  return (type);
}

int32_t SDDS_GetAssociateInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, ...) {
  int32_t field_index, type, associate_index;
  ASSOCIATE_DEFINITION *associatedef;
  char *associate_name;
  va_list argptr;
  int32_t retval;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_GetAssociateInformation"))
    return (0);

  if (!field_name) {
    SDDS_SetError("NULL field name passed. (SDDS_GetAssociateInformation)");
    return (0);
  }

  va_start(argptr, mode);
  retval = 1;
  if (mode & SDDS_GET_BY_INDEX) {
    if ((associate_index = va_arg(argptr, int32_t)) < 0 || associate_index >= SDDS_dataset->layout.n_associates) {
      SDDS_SetError("Invalid associate index passed. (SDDS_GetAssociateInformation)");
      retval = 0;
    }
  } else {
    if (!(associate_name = va_arg(argptr, char *))) {
      SDDS_SetError("NULL associate name passed. (SDDS_GetAssociateInformation)");
      retval = 0;
    }
    if ((associate_index = SDDS_GetAssociateIndex(SDDS_dataset, associate_name)) < 0) {
      SDDS_SetError("Unknown associate name given (SDDS_GetAssociateInformation)");
      retval = 0;
    }
  }
  associatedef = SDDS_dataset->layout.associate_definition + associate_index;
  va_end(argptr);
  if (!retval)
    return (0);

  for (field_index = 0; field_index < SDDS_ASSOCIATE_FIELDS; field_index++)
    if (strcmp(field_name, SDDS_AssociateFieldInformation[field_index].name) == 0)
      break;
  if (field_index == SDDS_ASSOCIATE_FIELDS) {
    SDDS_SetError("Unknown field name given (SDDS_GetAssociateInformation)");
    return (0);
  }
  type = SDDS_AssociateFieldInformation[field_index].type;
  if (!memory)
    return (type);
  if (type == SDDS_STRING) {
    if (!SDDS_CopyString((char **)memory, *((char **)((char *)associatedef + SDDS_AssociateFieldInformation[field_index].offset)))) {
      SDDS_SetError("Unable to copy field data (SDDS_GetAssociateInformation)");
      return (0);
    }
  } else
    memcpy(memory, (char *)associatedef + SDDS_AssociateFieldInformation[field_index].offset, SDDS_type_size[type - 1]);
  return (type);
}

/**
 * Used to change a column definition field
 *
 * @param SDDS_dataset: Address of the SDDS_DATASET structure for the data set.
 * @param field_name: A NULL terminated string giving the name of the field to be changed.
 * @param memory: The address with the new field value.
 * @param mode: A bit-wise combination of the constants: SDDS_SET_BY_INDEX, SDDS_SET_BY_NAME, SDDS_PASS_BY_VALUE and SDDS_PASS_BY_STRING\n
 *              The syntax for the four possible combinations is:\n
 *              SDDS_SET_BY_INDEX + SDDS_PASS_BY_VALUE: int32_t SDDS_ChangeColumnInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, int32_t column_index)\n
 *              SDDS_SET_BY_NAME + SDDS_PASS_BY_VALUE: int32_t SDDS_ChangeColumnInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, char *column_name)\n
 *              SDDS_SET_BY_INDEX + SDDS_PASS_BY_STRING: int32_t SDDS_ChangeColumnInformation(SDDS_DATASET *SDDS_dataset, char *field_name, char *memory, int32_t mode, int32_t column_index)\n
 *              SDDS_SET_BY_NAME + SDDS_PASS_B_YSTRING: int32_t SDDS_ChangeColumnInformation(SDDS_DATASET *SDDS_dataset, char *field_name, char *memory, int32_t mode, char *column_name)
 *
 * @return On success, returns the SDDS data type of the information. On failure, returns zero and records an error message.
 *
 */
int32_t SDDS_ChangeColumnInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, ...) {
  int32_t field_index, type, givenType;
  int32_t i, column_index;
  COLUMN_DEFINITION *columndef;
  char *column_name;
  va_list argptr;
  int32_t retval;
  double buffer[4];

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_ChangeColumnInformation"))
    return (0);

  if (!field_name) {
    SDDS_SetError("NULL field name passed. (SDDS_ChangeColumnInformation)");
    return (0);
  }

  va_start(argptr, mode);
  retval = 1;
  if (mode & SDDS_SET_BY_INDEX) {
    if ((column_index = va_arg(argptr, int32_t)) < 0 || column_index >= SDDS_dataset->layout.n_columns) {
      SDDS_SetError("Invalid column index passed. (SDDS_ChangeColumnInformation)");
      retval = 0;
    }
  } else {
    if (!(column_name = va_arg(argptr, char *))) {
      SDDS_SetError("NULL column name passed. (SDDS_ChangeColumnInformation)");
      retval = 0;
    }
    if ((column_index = SDDS_GetColumnIndex(SDDS_dataset, column_name)) < 0) {
      SDDS_SetError("Unknown column name given (SDDS_ChangeColumnInformation)");
      retval = 0;
    }
  }
  columndef = SDDS_dataset->layout.column_definition + column_index;
  va_end(argptr);
  if (!retval)
    return (0);

  for (field_index = 0; field_index < SDDS_COLUMN_FIELDS; field_index++)
    if (strcmp(field_name, SDDS_ColumnFieldInformation[field_index].name) == 0)
      break;
  if (field_index == SDDS_COLUMN_FIELDS) {
    SDDS_SetError("Unknown field name given (SDDS_ChangeColumnInformation)");
    return (0);
  }
  type = SDDS_ColumnFieldInformation[field_index].type;
  if (!memory)
    return (type);
  if (type == SDDS_STRING) {
    if (!SDDS_CopyString(((char **)((char *)columndef + SDDS_ColumnFieldInformation[field_index].offset)), (char *)memory)) {
      SDDS_SetError("Unable to copy field data (SDDS_ChangeColumnInformation)");
      return (0);
    }
    if (strcmp(field_name, "name") == 0) {
      for (i = 0; i < SDDS_dataset->layout.n_columns; i++)
        if (column_index == SDDS_dataset->layout.column_index[i]->index)
          break;
      if (i == SDDS_dataset->layout.n_columns) {
        SDDS_SetError("Unable to copy field data--column indexing problem (SDDS_ChangeColumnInformation)");
        return (0);
      }
      SDDS_dataset->layout.column_index[i]->name = SDDS_dataset->layout.column_definition[column_index].name;
      qsort((char *)SDDS_dataset->layout.column_index, SDDS_dataset->layout.n_columns, sizeof(*SDDS_dataset->layout.column_index), SDDS_CompareIndexedNamesPtr);
    }
  } else {
    if (mode & SDDS_PASS_BY_STRING) {
      if (strcmp(field_name, "type") == 0 && (givenType = SDDS_IdentifyType((char *)memory)) > 0)
        /* the type has been passed as a string (e.g., "double") */
        memcpy((char *)buffer, (char *)&givenType, sizeof(givenType));
      else if (!SDDS_ScanData((char *)memory, type, 0, (void *)buffer, 0, 0)) {
        SDDS_SetError("Unable to scan string data (SDDS_ChangeColumnInformation)");
        return (0);
      }
      memcpy((char *)columndef + SDDS_ColumnFieldInformation[field_index].offset, (void *)buffer, SDDS_type_size[type - 1]);
    } else
      memcpy((char *)columndef + SDDS_ColumnFieldInformation[field_index].offset, memory, SDDS_type_size[type - 1]);
  }
  return (type);
}

/**
 * Used to change a parameter definition field
 *
 * @param SDDS_dataset: Address of the SDDS_DATASET structure for the data set.
 * @param field_name: A NULL terminated string giving the name of the field to be changed.
 * @param memory: The address with the new field value.
 * @param mode: A bit-wise combination of the constants: SDDS_SET_BY_INDEX, SDDS_SET_BY_NAME, SDDS_PASS_BY_VALUE and SDDS_PASS_BY_STRING\n
 *              The syntax for the four possible combinations is:\n
 *              SDDS_SET_BY_INDEX + SDDS_PASS_BY_VALUE: int32_t SDDS_ChangeParameterInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, int32_t parameter_index)\n
 *              SDDS_SET_BY_NAME + SDDS_PASS_BY_VALUE: int32_t SDDS_ChangeParameterInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, char *parameter_name)\n
 *              SDDS_SET_BY_INDEX + SDDS_PASS_BY_STRING: int32_t SDDS_ChangeParameterInformation(SDDS_DATASET *SDDS_dataset, char *field_name, char *memory, int32_t mode, int32_t parameter_index)\n
 *              SDDS_SET_BY_NAME + SDDS_PASS_B_YSTRING: int32_t SDDS_ChangeParameterInformation(SDDS_DATASET *SDDS_dataset, char *field_name, char *memory, int32_t mode, char *parameter_name)
 *
 * @return On success, returns the SDDS data type of the information. On failure, returns zero and records an error message.
 *
 */
int32_t SDDS_ChangeParameterInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, ...) {
  int32_t field_index, type, parameter_index, givenType;
  PARAMETER_DEFINITION *parameterdef;
  char *parameter_name;
  va_list argptr;
  int32_t retval;
  double buffer[4];

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_ChangeParameterInformation"))
    return (0);

  if (!field_name) {
    SDDS_SetError("NULL field name passed. (SDDS_ChangeParameterInformation)");
    return (0);
  }

  va_start(argptr, mode);
  retval = 1;
  if (mode & SDDS_SET_BY_INDEX) {
    if ((parameter_index = va_arg(argptr, int32_t)) < 0 || parameter_index >= SDDS_dataset->layout.n_parameters) {
      SDDS_SetError("Invalid parameter index passed. (SDDS_ChangeParameterInformation)");
      retval = 0;
    }
  } else {
    if (!(parameter_name = va_arg(argptr, char *))) {
      SDDS_SetError("NULL parameter name passed. (SDDS_ChangeParameterInformation)");
      retval = 0;
    }
    if ((parameter_index = SDDS_GetParameterIndex(SDDS_dataset, parameter_name)) < 0) {
      SDDS_SetError("Unknown parameter name given (SDDS_ChangeParameterInformation)");
      retval = 0;
    }
  }
  parameterdef = SDDS_dataset->layout.parameter_definition + parameter_index;
  va_end(argptr);
  if (!retval)
    return (0);

  for (field_index = 0; field_index < SDDS_PARAMETER_FIELDS; field_index++)
    if (strcmp(field_name, SDDS_ParameterFieldInformation[field_index].name) == 0)
      break;
  if (field_index == SDDS_PARAMETER_FIELDS) {
    SDDS_SetError("Unknown field name given (SDDS_ChangeParameterInformation)");
    return (0);
  }
  type = SDDS_ParameterFieldInformation[field_index].type;
  if (!memory)
    return (type);
  if (type == SDDS_STRING) {
    if (!SDDS_CopyString(((char **)((char *)parameterdef + SDDS_ParameterFieldInformation[field_index].offset)), (char *)memory)) {
      SDDS_SetError("Unable to copy field data (SDDS_ChangeParameterInformation)");
      return (0);
    }
    if (strcmp(field_name, "name") == 0)
      qsort((char *)SDDS_dataset->layout.parameter_index, SDDS_dataset->layout.n_parameters, sizeof(*SDDS_dataset->layout.parameter_index), SDDS_CompareIndexedNamesPtr);
  } else {
    if (mode & SDDS_PASS_BY_STRING) {
      if (strcmp(field_name, "type") == 0 && (givenType = SDDS_IdentifyType((char *)memory)) > 0)
        /* the type has been passed as a string (e.g., "double") */
        memcpy((char *)buffer, (char *)&givenType, sizeof(givenType));
      else if (!SDDS_ScanData((char *)memory, type, 0, (void *)buffer, 0, 0)) {
        SDDS_SetError("Unable to scan string data (SDDS_ChangeParameterInformation)");
        return (0);
      }
      memcpy((char *)parameterdef + SDDS_ParameterFieldInformation[field_index].offset, (void *)buffer, SDDS_type_size[type - 1]);
    } else
      memcpy((char *)parameterdef + SDDS_ParameterFieldInformation[field_index].offset, memory, SDDS_type_size[type - 1]);
  }

  return (type);
}

/**
 * Used to change a array definition field
 *
 * @param SDDS_dataset: Address of the SDDS_DATASET structure for the data set.
 * @param field_name: A NULL terminated string giving the name of the field to be changed.
 * @param memory: The address with the new field value.
 * @param mode: A bit-wise combination of the constants: SDDS_SET_BY_INDEX, SDDS_SET_BY_NAME, SDDS_PASS_BY_VALUE and SDDS_PASS_BY_STRING\n
 *              The syntax for the four possible combinations is:\n
 *              SDDS_SET_BY_INDEX + SDDS_PASS_BY_VALUE: int32_t SDDS_ChangeArrayInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, int32_t array_index)\n
 *              SDDS_SET_BY_NAME + SDDS_PASS_BY_VALUE: int32_t SDDS_ChangeArrayInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, char *array_name)\n
 *              SDDS_SET_BY_INDEX + SDDS_PASS_BY_STRING: int32_t SDDS_ChangeArrayInformation(SDDS_DATASET *SDDS_dataset, char *field_name, char *memory, int32_t mode, int32_t array_index)\n
 *              SDDS_SET_BY_NAME + SDDS_PASS_B_YSTRING: int32_t SDDS_ChangeArrayInformation(SDDS_DATASET *SDDS_dataset, char *field_name, char *memory, int32_t mode, char *array_name)
 *
 * @return On success, returns the SDDS data type of the information. On failure, returns zero and records an error message.
 *
 */
int32_t SDDS_ChangeArrayInformation(SDDS_DATASET *SDDS_dataset, char *field_name, void *memory, int32_t mode, ...) {
  int32_t field_index, type, array_index, givenType;
  ARRAY_DEFINITION *arraydef;
  char *array_name;
  va_list argptr;
  int32_t retval;
  double buffer[4];

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_ChangeArrayInformation"))
    return (0);

  if (!field_name) {
    SDDS_SetError("NULL field name passed. (SDDS_ChangeArrayInformation)");
    return (0);
  }

  va_start(argptr, mode);
  retval = 1;
  if (mode & SDDS_SET_BY_INDEX) {
    if ((array_index = va_arg(argptr, int32_t)) < 0 || array_index >= SDDS_dataset->layout.n_arrays) {
      SDDS_SetError("Invalid array index passed. (SDDS_ChangeArrayInformation)");
      retval = 0;
    }
  } else {
    if (!(array_name = va_arg(argptr, char *))) {
      SDDS_SetError("NULL array name passed. (SDDS_ChangeArrayInformation)");
      retval = 0;
    }
    if ((array_index = SDDS_GetArrayIndex(SDDS_dataset, array_name)) < 0) {
      SDDS_SetError("Unknown array name given (SDDS_ChangeArrayInformation)");
      retval = 0;
    }
  }
  arraydef = SDDS_dataset->layout.array_definition + array_index;
  va_end(argptr);
  if (!retval)
    return (0);

  for (field_index = 0; field_index < SDDS_ARRAY_FIELDS; field_index++)
    if (strcmp(field_name, SDDS_ArrayFieldInformation[field_index].name) == 0)
      break;
  if (field_index == SDDS_ARRAY_FIELDS) {
    SDDS_SetError("Unknown field name given (SDDS_ChangeArrayInformation)");
    return (0);
  }
  type = SDDS_ArrayFieldInformation[field_index].type;
  if (!memory)
    return (type);
  if (type == SDDS_STRING) {
    if (!SDDS_CopyString(((char **)((char *)arraydef + SDDS_ArrayFieldInformation[field_index].offset)), (char *)memory)) {
      SDDS_SetError("Unable to copy field data (SDDS_ChangeArrayInformation)");
      return (0);
    }
    if (strcmp(field_name, "name") == 0)
      qsort((char *)SDDS_dataset->layout.array_index, SDDS_dataset->layout.n_arrays, sizeof(*SDDS_dataset->layout.array_index), SDDS_CompareIndexedNamesPtr);
  } else {
    if (mode & SDDS_PASS_BY_STRING) {
      if (strcmp(field_name, "type") == 0 && (givenType = SDDS_IdentifyType((char *)memory)) > 0)
        /* the type has been passed as a string (e.g., "double") */
        memcpy((char *)buffer, (char *)&givenType, sizeof(givenType));
      else if (!SDDS_ScanData((char *)memory, type, 0, (void *)buffer, 0, 0)) {
        SDDS_SetError("Unable to scan string data (SDDS_ChangeArrayInformation)");
        return (0);
      }
      memcpy((char *)arraydef + SDDS_ArrayFieldInformation[field_index].offset, (void *)buffer, SDDS_type_size[type - 1]);
    } else
      memcpy((char *)arraydef + SDDS_ArrayFieldInformation[field_index].offset, memory, SDDS_type_size[type - 1]);
  }

  return (type);
}
